from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#!/usr/bin/env python

import sys
from . import funcs_main as fmain
from . import symT
from . import input
from . import fslib
from .version import __version__

if '--version' in sys.argv:
    print('Symmetr version: {}'.format(__version__))
    print('Python version: {}'.format(sys.version))

opt = input.parse()

if opt['print_syms']:
    syms = symT.get_syms(opt)
    print('Symmetry operations:')
    print('Format: Number, space transformation, magnetic moment transformation, time-reversal, transformation of the sublattices')
    for i,sym in enumerate(syms):
        print('Symmetry: ', i)
        print(sym)
    if opt['mode'] == 'res' and opt['noso']:
        syms_noso = symT.get_syms_noso(opt)
        print('')
        print('Noso symmetry operations (in the magnetic basis)')
        for i,sym in enumerate(syms_noso):
            print(i+1,sym)

if opt['print_opt']:
    print(opt)

if opt['print_pos']:
    lines = fslib.run_fs(opt['inp'])
    pos = fslib.r_pos(lines)
    print('Atomic sites as used by findsym:')
    for p in pos:
        print(p[-1],p[0:3])
    print('')

if opt['mode'] == 'res':
    fmain.sym_res(opt,printit=True)
else:
    fmain.sym_mham(opt,printit=True) 



from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import str
from builtins import range
import re
import copy
import sys
import math
import time

import sympy
import numpy as np

from .tensors import matrix, mat2ten, tensor
from .fslib import transform_position
from .symmetrize import symmetr,params_trans
from .symT import convert_vec

def create_rank2(ten,n_op=2,xyz=False):
    """
    Creates a rank 2 tensor that includes magnetic moment explicitely.
    """

    X = matrix(0,3)
    X.x = ten.x
    X.v = ten.v

    if ten.dim2 > 2:

        m = {}
        for i in range(3):
            if xyz:
                names = ['m_x','m_y','m_z']
                name = names[i]
            else:
                name = 'm%s' % i
            m[i] = sympy.symbols(name)


        for ind in ten:
            M = 1
            for i in range(2,len(ind)):
                M *= m[ind[i]]
            X[ind[0],ind[1]] += ten[ind]*M


    else:
        for ind in ten:
            X[ind] = ten[ind]
    return X

def sub_m(ten,n_op,xyz=False):
    
    X = tensor(0,3,n_op)

    if ten.dim2 > n_op:
        m = {}
        for i in range(3):
            if xyz:
                names = ['m_x','m_y','m_z']
                name = names[i]
            else:
                name = 'm%s' % i
            m[i] = sympy.symbols(name)
        for ind in ten:
            M = 1
            for i in range(n_op,len(ind)):
                M *= m[ind[i]]
            ind_m = []
            for i in range(n_op):
                ind_m.append(ind[i])
            ind_m = tuple(ind_m)
            X[ind_m] += ten[ind]*M

    else:
        for ind in ten:
            X[ind] = ten[ind]

    return X

def print_tensor(ten,n_op,latex=False,xyz=False,no_newline=False):
    """
    Prints the expansion tensor in a nice form.

    Not tested for higher order than 1!!!
    """
    
    X = sub_m(ten,n_op,xyz=xyz)

    if not latex:
        X.pprint()
    else:
        X.pprint(latex=True,no_newline=no_newline)

def simplify_tensor(ten,xyz=False,index_from_1=False):
    """
    Renames the variables of the tensor and simplifies it.
    """

    X = create_rank2(ten,xyz=xyz)
    xinds = list(set(re.findall(r'x[0-9]+',sympy.srepr(X))))

    #contains the new indices
    xn = {}
    if not index_from_1:
        for i in range(len(xinds)):
            xn[i] = sympy.symbols('x'+str(i))
    else:
        for i in range(len(xinds)):
            xn[i] = sympy.symbols('x'+str(i+1))

    for ind in X:
        for i in range(len(xinds)):
            X[ind] = X[ind].subs(xinds[i],xn[i])
        X[ind] = sympy.simplify(X[ind])

    return X

def index_from_1(X,rank=2):
    """
    Takes a rank 3 tensor and rename the indeces so that the numbering starts from 1 and not 0.
    """

    if rank == 2:
    
        xinds = list(set(re.findall(r'x[0-9]+',sympy.srepr(X))))
        xn = {}
        xnz = {}
        for i in range(3):
            for j in range(3):
                xnz[(i,j)] = sympy.symbols('z'+str(i+1)+str(j+1))

        for i in range(3):
            for j in range(3):
                xn[(i,j)] = sympy.symbols('x'+str(i+1)+str(j+1))

        for ind in X:
            for i in range(3):
                for j in range(3):
                    X[ind] = X[ind].subs(X.x[i,j],xnz[(i,j)])

        for ind in X:
            for i in range(3):
                for j in range(3):
                    X[ind] = X[ind].subs(xnz[(i,j)],xn[(i,j)])

        return X

    if rank == 3:

        xinds = list(set(re.findall(r'x[0-9]+',sympy.srepr(X))))
        xn = {}
        xnz = {}
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    xnz[(i,j,k)] = sympy.symbols('z'+str(i+1)+str(j+1)+str(k+1))

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    xn[(i,j,k)] = sympy.symbols('x'+str(i+1)+str(j+1)+str(k+1))

        for ind in X:
            for i in range(3):
                for j in range(3):
                    for k in range(3):
                        X[ind] = X[ind].subs(X.x[i,j,k],xnz[(i,j,k)])

        for ind in X:
            for i in range(3):
                for j in range(3):
                    for k in range(3):
                        X[ind] = X[ind].subs(xnz[(i,j,k)],xn[(i,j,k)])

        return X

def is_collinear(mags,prec=1e-5):
    res = True
    first = sorted(list(mags.keys()))[0]
    for i in mags:
        if abs(mags[first].dot(mags[i])) < 1-prec:
            res = False
            break

    return res

def convert_mags(mags,sym):
    mags_R = {}
    for atom in mags:
        mags_R[atom] = convert_vec(mags[atom],sym.get_R('s'))
    mags_Rp = {}
    perms = sym.permutations
    for atom in mags:
        mags_Rp[perms[atom]] = mags_R[atom]

    return mags_Rp

def get_L_trans(mags,sym,debug=False):

    initial_signs = []
    
    first = sorted(list(mags.keys()))[0]
    for atom in mags:
        if mags[first].dot(mags[atom]) > 0:
            initial_signs.append(1)
        else:
            initial_signs.append(-1)

    mags_R = convert_mags(mags,sym)
    signs = []
    for atom in mags_R:
        if mags_R[first].dot(mags_R[atom]) > 0:
            signs.append(1)
        else:
            signs.append(-1)

    if debug:
        print('')
        print('transformed magnetic moments:')
        for atom in  sorted(list(mags_R.keys())):
            print(atom, ":")
            sympy.pprint(mags_R[atom])

        print('initial signs of magnetic moments')
        print(initial_signs)

        print('signs of the trasnformed moments')
        print(signs)

    if initial_signs != signs:
        return None
    else:
        mag_0_R = convert_vec(mags[first],sym.get_R('s'))
        sign_0 = mag_0_R.dot(mags[first])
        if sign_0 > 0:
            sign = 1
        else:
            sign = -1
        return sign * sym.get_R('s')

def def_syms_L(mags,syms,prec=1e-5,debug=False):

    #select nonzero magnetic moments
    mags_dict = {}
    for i,mag in enumerate(mags):
        if mag.norm() > prec:
            mags_dict[i+1] = mag

    if debug:
        print('Starting defining L transformation')
        print('Nonzero magnetic moments')
        print(mags,mags_dict)
        for atom in  sorted(list(mags_dict.keys())):
            print(atom, ":")
            sympy.pprint(mags_dict[atom])

    if len(mags_dict) == 0:
        print("!!!Warning!!!: no magnetic moments defined in the input. Assuming a ferromagnetic system.")
        for sym in syms:
            sym.def_custom_R('L',sym.get_R('s'))
        syms_L = syms

    else:

        if not is_collinear(mags_dict,prec):
            raise Exception("Expansions only work for collinear magnetic systems")

        syms_L = []
        for sym in syms:

            if debug:
                print('')
                print('Taking symmetry:')
                print(sym)

            L_trans = get_L_trans(mags_dict,sym,debug=debug)
            if debug:
                print('L_trans:')
                print(L_trans)
            if L_trans is not None:
                sym.def_custom_R('L',L_trans)
                syms_L.append(sym)

    return syms_L







            








# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Defines a symbolic tensor class.

All the tensors that are symmetrized use this class.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
from  six import string_types

import sympy
from sympy.core.numbers import Zero
import numpy as np
import copy
from .symmetry import create_T
import prettytable
from prettytable import PrettyTable

class tensor(object):
    """
    Creates a symbolic tensor.

    Usage:
        Use by t=tensor('s',dim1,dim2), where 's' means the tensor should be symbolic, dim2 is the number of indeces
             of the tensor, dim1 is the number of values each index can have.
        By default variables will be called x### for a different variable name, Run with tensor('s',dim1,dim2,'name').
        For nonsymbolic zero tensor, run  tensor(0,dim1,dim2).
        Print by print t
        Write out element of the tensor by t[1,2,3] or t[(1,2,3)] for example
        Assign a value to element: t[1,2,3] = 0.
        Sum or substract two tensors: t1+t2.
        Multiply by a number or symbolic variable: t*2.
        Loop over all elements, following will print all elements for example. Note that the loop is over indeces, not elements!
         for i in t:
           print t[i]
        If you need to use the symbolic variables use t.x[index], for example for x011, use t.x[0,1,1] (or t.x[(0,1,1)]).
            You can do for example t[0,0,0] = t.x[1,1,1], then t[0,0,0] will be equal to x111.
            You can also access x011 by t['x011'].
    """

    def __init__(self,kind,dim1,dim2,name='x',ind_types=None):
        """
        Creates a symbolic tensor.

        Args:
            kind : Sets what kind of tensor. 's' for symbolic, 0 for zero tensor
            dim1 (int): How many values can each index have.
            dim2 (int): How many indeces are there in the tensor.
            name (optional[str]): Name of the symbolic variables. Defaults to 'x'.
            ind_types (optinal[tuple]): Specifies whether individual indeces are covariant
                or contravariant. 1 specifies contravariant and -1 covariant. If not specified
                all the indeces are assumed contravariant.
        """

        self.inds = makeinds(dim1,dim2)
        self.dim1 = dim1
        self.dim2 = dim2
        self.name = name
        if ind_types is None:
            self.ind_types = (1,)*self.dim2
        else:
            self.ind_types = ind_types
        self.trans_type = None
        if kind == 's':
            self.v = {}
            for ind in self.inds:
                s_tot = var_name(name,ind,self.dim1)
                self.v[s_tot] = sympy.symbols(s_tot)
        # self.t contains the tensor itself, stored as a dictionary
        self.t = {}
        #self.x contains the symbolic variables in a form that can be easily retrieved
        self.x = {}
        type_found = False
        for ind in self.inds:
            if kind == 0:
                type_found = True
                self.t[ind] = 0
            if kind == 's':
                type_found = True
                n_ind = var_name(name,ind,self.dim1)
                self.t[ind] = self.v[n_ind]
                self.x[ind] = self.v[n_ind]
        if type_found == False:
            sys.exit('wrong tensor type')

        self.G = None
        self.Gi = None

    def __getitem__(self,key):
        if type(key) == tuple:
            return self.t[key]
        elif isinstance(key,int):
            return self.t[(key,)]
        elif isinstance(key,string_types):
            return self.v[key]
        raise Exception('Unexpected key type for key {} with type {}.'.format(key,type(key)))
    def __setitem__(self,key,value):
        if type(key) != tuple:
            raise TypeError
        elif len(key) != self.dim2:
            raise TypeError
        else:
            for i in key:
                if i not in list(range(self.dim1)):
                    raise TypeError
        self.t[key]=value
    def __len__(self):
        return len(self.inds)
    def __add__(self,other):
        out = self.copy0()
        for ind in self.inds:
            out[ind] = self[ind] + other[ind]
        return out
    def __sub__(self,other):
        out = self.copy0()
        for ind in self.inds:
            out[ind] = self[ind] - other[ind]
        return out
    def __mul__(self,num):
        out = self.copy0()
        for ind in self.inds:
            out[ind] = self[ind] * num
        return out
    def __div__(self,num):
        out = self.copy0()
        for ind in self.inds:
            out[ind] = old_div(self[ind], num)
        return out
    def __radd__(self,other):
        return self + other
    def __rsub__(self,other):
        return self - other
    def __rmul__(self,num):
        return self*num
    def __neg__(self):
        return self*-1
    def __str__(self):
        out = ''
        for ind in self:
            out = out + ind.__str__() + ' ' + self[ind].__str__() + '\n'
        return out
    def __iter__(self):
        return iter(self.inds)
    def __eq__(self,other):
        out = True
        if self.dim1 != other.dim1 or self.dim2 != other.dim2:
            out = False
        else:
            for ind in self.inds:
                if sympy.simplify(self[ind]-other[ind]) != 0:
                    out = False
        return out
    def __ne__(self,other):
        if self == other:
            return False
        else:
            return True
    def __type__(self):
        return 'tensor'

    def mat(self,numpy=False):
       #outputs a matrix form of itself either in sympy format (default) or numpy format 
       if self.dim2 != 2:
           raise TypeError
       if not numpy:
           out = sympy.zeros(self.dim1,self.dim1)
       if numpy:
           out = np.zeros((self.dim1,self.dim1))
       for i in range(self.dim1):
           for j in range(self.dim1):
               out[i,j] = self[i,j]
       return out

    def subs(self,old,new='notset'):
        for ind in self:
            if new != 'notset':
                self[ind] = self[ind].subs(old,new)
            else:
                self[ind] = self[ind].subs(old)
        return self
    def reduce(self,comp,value):
        out = tensor(0,self.dim1,self.dim2-1)
        for ind in out:
            ind2 = [0]*self.dim2
            for i in range(self.dim2):
                if i < comp:
                    ind2[i] = ind[i]
                if i == comp:
                    ind2[i] = value
                if i > comp:
                    ind2[i] = ind[i-1]
            ind2 = tuple(ind2)
            out[ind] = self[ind2]
        return out

    def pprint(self,print_format=None,latex=False,no_newline=False,ind_types=False,remove_zeros=False):

        if remove_zeros:
            self.remove_zeros()
        if ind_types:
            print(self.ind_types)
        if self.dim2 == 1:
            vec = [self[i] for i in range(self.dim1)]
            vec = sympy.Matrix([vec])
            if latex:
                print(sympy.latex(vec))
            else:
                sympy.pprint(vec)
        elif self.dim2 == 2:
            if latex:
                if no_newline:
                    print(sympy.latex(self.mat()), end=' ')
                else:
                    print(sympy.latex(self.mat()))
            else:
                sympy.pprint(self.mat())
        else:

            if print_format is None:
                if self.dim2 == 3:
                    print_format = 1
                else:
                    print_format = 2

            if print_format == 0:

                print(self)

            elif print_format == 1:

                if self.dim2 == 3:
                    print('X_ijk' + ' =')
                elif self.dim2 == 4:
                    print('X_ijkl' + ' =')
                else:
                    print('X_ij...pq' + ' =')

                r_inds = makeinds(self.dim1,self.dim2-2)
                for r_ind in r_inds:
                    out_str = 'X_'
                    for i in range(len(r_ind)):
                        if i == 0:
                            X = self.reduce(0,r_ind[i])
                        else:
                            X = X.reduce(0,r_ind[i])
                        #if i > 0:
                        #    out_str += ','
                        out_str += '{0}'.format(r_ind[i])
                    out_str += 'pq ='
                    print(out_str) 
                    if latex:
                        print(sympy.latex(X.mat()))
                    else:
                        X.pprint()

            elif print_format == 2:

                if latex:
                    print('WARNING: latex format not supported for print_format = 2, use print_format = 1')
                    return None

                r_inds = makeinds(self.dim1,self.dim2-2)
                r_inds2 = makeinds(self.dim1,2)
                t = PrettyTable()
                if self.dim2 == 3:
                    ind_names = ['i','j','k']
                elif self.dim2 == 4:
                    ind_names = ['i','j','k','l']
                else:
                    ind_names = ['i','j','p','q']

                t.title = '({0},{1})'.format(ind_names[-2],ind_names[-1])
                t.field_names = ([""]+[" "]+r_inds2)
                n_r_inds = len(r_inds)
                for i,r_ind in enumerate(r_inds):
                    row = [r_ind]
                    for r_ind2 in r_inds2:
                        ind = r_ind + r_ind2
                        row.append(self[ind])
                    if i == int(old_div(n_r_inds,2)):
                        if self.dim2 == 3:
                            label = "({0})".format(ind_names[0])
                        elif self.dim2 == 4:
                            label = "({0},{1})".format(ind_names[0],ind_names[1])
                        else:
                            label = "({0},{1},...)".format(ind_names[0],ind_names[1])

                    else:
                        label = ""
                    t.add_row([label]+row)
                t.hrules = prettytable.ALL
                print(t)
            else:
                raise Exception('print_format not recognized')

    def simplify(self):
        for ind in self:
            self[ind] = sympy.simplify(self[ind])

    def remove_zeros(self,num_digits=14):
        for ind in self:
            symbols = self[ind].free_symbols
            for symbol in symbols:
                if abs(self[ind].coeff(symbol)) < 1e-6:
                    self[ind] = self[ind].subs(symbol,0)

    def evalf(self,acc=15):
        out = 0

    def copy0(self):
        "returns a zero tensor with the same shape and transformation rules"
        out = tensor(0,self.dim1,self.dim2,ind_types=self.ind_types)
        if self.trans_type is not None:
            out.def_trans(ind_trans=self.ind_trans,T_comp=self.T_comp,P_trans=self.P_trans,T_trans=self.T_trans)
        if self.G is not None:
            out.def_metric(self.G)
        return out

    def copy(self):
        "returns a copy of itself"
        out = self.copy0()
        for ind in self:
            out[ind] = self[ind]
        return out

    def is_contravar(self,i):
        """Checkes if the index i is contravariant."""
        if self.ind_types[i] == 1:
            return True
        elif self.ind_types[i] == -1:
            return False
        else:
            raise Exception('Wrong ind type')

    def is_covar(self,i):
        """Checkes if the index i is covariant."""
        if self.ind_types[i] == 1:
            return False
        elif self.ind_types[i] == -1:
            return True
        else:
            raise Exception('Wrong ind type')

    def convert(self,T,in_place=True):
        """
        Return the tensor transormed by coordinate transformation matrix T.

        Args:
            T (matrix): Coordinate transformation matrix. T transforms from A to B, ie Tx_A = x_B.

        Returns:
            ten_T (tensor): the transformed tensor
        """
        
        mat1 = T
        mat2 = T.inv().T

        ten_T = self.copy0()
        for ind1 in ten_T:
            for ind2 in self:
                factor = 1
                for i in range(self.dim2):
                    if self.is_contravar(i):
                        factor *= mat1[ind1[i],ind2[i]]
                    else:
                        factor *= mat2[ind1[i],ind2[i]]
                ten_T[ind1] += factor*self[ind2]

        if in_place:
            for ind in self:
                self[ind] = ten_T[ind]

        if self.G is not None:
            G_T = T * self.G * T.T
            Gi_T = T.T.inv() * self.Gi * T.inv()
            if in_place:
                self.G = G_T 
                self.Gi = Gi_T
            else:
                ten_T.G = G_T
                ten_T.Gi = Gi_T
        
        if not in_place:
            return ten_T

    def def_trans(self,ind_trans=None,T_comp=None,P_trans=None,T_trans=None):

        inp_type_1 =  ind_trans is not None and T_comp is not None
        inp_type_2 =  P_trans is not None and T_trans is not None
        if inp_type_1 and inp_type_2:
            raise Exception('You have to specify either ind_trans and T_comp or P_trans and T_trans')
        if (not inp_type_1) and (not inp_type_2):
            raise Exception('You have to specify either ind_trans and T_comp or P_trans and T_trans')

        if inp_type_1:
            self.trans_type = 1
        if inp_type_2:
            self.trans_type = 2
        self.ind_trans = ind_trans
        self.T_comp = T_comp
        self.P_trans = P_trans
        self.T_trans = T_trans

    def transform(self,sym):

        if self.trans_type is None:
            raise Exception('To transform the tensor, the transformation rules must be defined')

        R_list = []
        for i in range(self.dim2):
            if self.trans_type == 1:
                R = sym.get_R(self.ind_trans[i])
            elif self.trans_type == 2:
                R = sym.R
            if self.is_contravar == 1:
                R_list.append(R)
            else:
                R_list.append(R.inv().T)

        ten_R = self.copy0()

        factor_ini = 1
        if self.trans_type == 1:
            if sym.has_T and self.T_comp == -1:
                factor_ini *= -1
        elif self.trans_type == 2:
            if sym.has_T and self.T_trans == -1:
                factor_ini *= -1
            if R_list[0].det() == -1 and self.P_trans == -1:
                factor_ini *= -1

        for ind1 in ten_R:
            for ind2 in self:
                factor = factor_ini
                for i in range(self.dim2):
                    factor *= R_list[i][ind1[i],ind2[i]]
                ten_R[ind1] += factor*self[ind2]

        return ten_R

    def is_even(self,sym=None):

        if self.trans_type is None:
            raise Exception('To transform the tensor, the transformation rules must be defined')

        if sym is None:
            T = create_T()
        else:
            T = sym
        X_T = self.transform(T)
        if X_T == self:
            return True
        elif X_T == -self:
            return False
        else:
            raise Exception('Wrong transformation under time-reversal')

    def def_metric(self,G):
        """
        Defines the metric for the coordinate system in which the tensor is defined

        G should be the covariant metric.
        """

        self.G = G
        self.Gi =G.inv()

    def reverse_index(self,i,in_place=True):

        if self.ind_types[i] == 1:
            if self.Gi is None:
                raise Exception("Metric tensor not defined")
            Gt = self.Gi
        if self.ind_types[i] == -1:
            if self.G is None:
                raise Exception("Metric tensor not defined")
            Gt = self.G

        out = self.copy0()

        for ind in self:
            out[ind] = 0
            for j in range(self.dim1):
                ind2 = list(ind)
                ind2[i] = j
                ind2 = tuple(ind2)
                out[ind] += Gt[ind[i],j] * self[ind2]

        if in_place:
            for ind in self:
                self[ind] = out[ind]

        ind_types = list(self.ind_types)
        ind_types[i] = -ind_types[i]
        if in_place:
            self.ind_types = tuple(ind_types)
        else:
            out.ind_types = tuple(ind_types)

        if not in_place:
            return out

    def raise_index(self,i):

        if self.ind_types[i] == 1:
            raise Exception("Index is already covariant")
        else:
            return self.reverse_index(i)

    def lower_index(self,i):

        if self.ind_types[i] == -1:
            raise Exception("Index is already contravariant")
        else:
            return self.reverse_index(i)

    def rename_vars(self,name=None,xyz=False):
        if name is None:
            name = self.name
        if xyz and self.dim1 != 3:
            raise Exception('xyz allowed only if dim1 == 3')

        new_vars = {} 

        for ind in self:
            if xyz:
                new_vars[ind] = sympy.symbols(var_name_xyz(name,ind))
            else:
                new_vars[ind] = sympy.symbols(var_name(name,ind,self.dim1))
        
        for ind in self:
            for ind2 in self:
                self[ind] = self[ind].subs(self.x[ind2],new_vars[ind2])

    def round(self,prec):
        def float2int(a):
            if abs(a) > 1e-14:
                if abs(old_div(int(a),a)-1) < 1e-14:
                    res = int(a)
                else:
                    res = a
            else:
                res = 0
            return res
        def round_expr(expr, num_digits):
            return expr.xreplace({n : float2int(round(n, num_digits)) for n in expr.atoms(sympy.Number)})
        for ind in self:
            if self[ind] != 0:
                self[ind] = round_expr(self[ind],prec)

class matrix(tensor):

     def __init__(self,kind,dim1,name='x'):
         tensor.__init__(self,kind,dim1,2,name)

     def __mul__(self,other):
         if isinstance(other, matrix):
                 out = mat2ten(self.mat()*other.mat())
         else:
             out = matrix(0,self.dim1)
             for ind in self.inds:
                 out[ind] = self[ind] * other
         return out

     def __rmul__(self,other):
         if isinstance(other, matrix):
             out = mat2ten(other.mat()*self.mat())
         else:
             out = matrix(0,self.dim1)
             for ind in self.inds:
                 out[ind] = self[ind] * other
         return out

     def __add__(self,other):
         out = matrix(0,self.dim1)
         for ind in self.inds:
             out[ind] = self[ind] + other[ind]
         return out

     def __sub__(self,other):
         out = matrix(0,self.dim1,self.dim2)
         for ind in self.inds:
             out[ind] = self[ind] - other[ind]
         return out

     def __div__(self,num):
         out = matrix(0,self.dim1,self.dim2)
         for ind in self.inds:
             out[ind] = old_div(self[ind], num)
         return out

     def __radd__(self,other):
         return self + other

     def __rsub__(self,other):
         return self - other

     def T(self):
         return mat2ten(self.mat().T)

    
            

def makeinds(dim1,dim2):

     for d in range(dim2):
         if d == 0:
             num = list(range(dim1))
         else:
             num1 = copy.deepcopy(num)
             num = []
             for n in num1:
                 for i in range(dim1):
                     if d == 1:
                         num.append([i]+[n])
                     else:
                         num.append([i]+n)
     for i in range(len(num)):
         if isinstance(num[i], (int,int)):
             num[i] = (num[i],)
         else:
             num[i] = tuple(num[i])

     return num


def var_name(name,num,dim):

     s_ind = ''
     for i in num:
         if dim < 11:
             s_ind = s_ind + str(i)
         else:
             s_ind = s_ind + '_' + str(i)

     if dim <11:
         s_tot = '%s%s' % (name,s_ind)
     else:
         s_tot = '%s%s' % (name,s_ind)

     return s_tot 

def var_name_xyz(name,num):
    xyz = ['x','y','z']
    s_ind = ''
    for i in num:
        s_ind = s_ind + xyz[i] 

    s_tot = name + '_' + s_ind

    return s_tot
def mat2ten(mat):
     try:
         ncols = mat.cols
         nrows = mat.rows
     except AttributeError:
         ncols = mat.shape[0]
         nrows = mat.shape[1]
     if ncols != nrows:
         raise TypeError
     else:
         out = matrix(0,ncols,2)
         for i in range(ncols):
             for j in range(ncols):
                 out[i,j] = mat[i,j]
         return out
     



from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import str
from builtins import range
import sys
import os
import re
import numpy as np
import sympy 
from .tensors import matrix
from .fslib import r_sym
from .symmetry import Symmetry,findsym2sym

dirname, filename = os.path.split(os.path.abspath(__file__))
sys.path.append(str(dirname))

def read_all_syms(hexag):
    """
    Generates a list of all crystallographic symmetry operations.

    Args:
        hexag(boolean): true if the conventional crystallographic group of the nonmagnetic crystal
            is hexagonal or trigonal, false otherwise.
    """

    if not hexag:
        syms_file = '/findsym/syms_table.dat'
    else:
        syms_file = '/findsym/syms_table_hex.dat'

    with open(dirname+syms_file) as f:
        lines = f.readlines()

    syms_list = []
    syms_list.append("_space_group_symop.magn_operation_mxmymz")
    for i,line in enumerate(lines):
        split = re.findall('[a-zA-Z0-9\-,\+]+',line)
        sym = str(i) + '  ' + split[1]+ ',+1  ' + split[2] 
        if not hexag:
            syms_list.append(sym)
        else:
            if 'T' not in line:
                syms_list.append(sym)


    syms_list.append(" ")
    syms = r_sym(syms_list,syms_only=True)
    mats = []
    for sym in syms:
        sym = findsym2sym(sym)
        mats.append(sym.get_R('x'))

    return mats

def noso_syms(syms,mag_conf,hexag,prec=1e-5,debug=False):
    """
    Takes symmetry operations in a nonmagnetic crystal and returns symmetry operations of a
        magnetic crystal without spin-orbit coupling.

    Args:
        syms: list of symmetry operations as returned by r_sym
        mag_conf([sympy(3)]): list of magnetic moments of every atom 
        hexag(boolean): true if the conventional crystallographic group of the nonmagnetic crystal
            is hexagonal or trigonal, false otherwise.

    Returns:
        syms_noso: list of symmetry operations without spin-orbit coupling. Format is different than 
            normal, the symmetry operations are in matrix form. Some functions can take argument
            sym_format='mat' to switch to this format.

    The algorithm:
        * Without SOC an arbitrary spin-rotation is a symmetry in a non-
            magnetic system. In a magnetic system,only spin-rotations that respect the magnetic order 
            are symmetries.
        * Thus we take every symmetry operation of the nonmagnetic system and try to see if we find some
            spin-rotations such that the symmetry+spin-rotation is a symmetry of the magnetic system.
        * The problem is that there is usually infinitely many such spin-rotations. It's not trivial to find them 
            in general and also to choose some finite number out of these.
        * Thus we use a trick. We do everything in the conventional coordinate basis because there the symmetry 
            operations have a simple form. We look for a combination of the symmetry operation+spin-rotation, which
            keeps the magnetic order invariant. We look at every crystallographic symmetry operation and see if it
            keeps the magnetic order invariant.
        * Rather than looking for the spin-rotatin we directly look for the combination of the symmetry +
            spin-rotation. Since the spin-rotation is arbitrary this is fine.
        * Since the spin-rotation is a proper rotation it must have determinant 1. Thus if there is time reversal,
            the total matrix (spin-rotation + the symmetry operation of the nonmagnetic crystal) must then have
            determinant -1. If there is no time-reversal in the symmetry operation it must have det +1. This is
            the only requirement.
        * In this way we get usually more than 1 symmetry operation for every nonmagnetic symmetry operation.
        * However, we do not get all the symmetry operations, in fact usually there is infinitely many of them.
        * This approach will probably work fine if the magnetic moments lie some high-symmetry directions. If they
            are not, then this approach will not find the correct spin-rotations and the symmetry will thus not
            be complete.
        * The symmetry should not be wrong, but can be incomplete - some elements can be shown as nonzero even
            though they should vanish or some relations missing.
    """

    if debug:
        if hexag:
            print('The conventional coordinate system is hexagonal.')
        else:
            print('The conventional coordinate system is not hexagonal.')
        print('')

    if debug:
        print('Magnetic moments are:')
        for i,mag in enumerate(mag_conf):
            print(i+1,' ', end=' ')
            sympy.pprint(mag)

    mats = read_all_syms(hexag)
    if debug:
        print('')
        print('list of all crystallographic symmetry operations:')
        for i,mat in enumerate(mats):
            print('')
            print(i+1)
            sympy.pprint(mat)
            print('')

    syms_noso = []
    start_new = 0
    for nsym,sym in enumerate(syms):
        start_new = len(syms_noso)

        if debug:
            print('')
            print('taking symmetry operation ', nsym+1)
            print('space part:')
            sympy.pprint(sym.R)
            print('magnetic part:')
            sympy.pprint(sym.Rs)
            print('time reversal:')
            sympy.pprint(sym.has_T)
            print('permutations:')
            print(sym.permutations)

        nmag = 0
        mag_is = []
        for i,mag in enumerate(mag_conf):
            if sympy.sqrt(mag[0]**2+mag[1]**2+mag[2]**2) > prec:
                nmag += 1
                mag_is.append(i)
        
        Y = sympy.zeros(nmag*3,10)

        for i in range(nmag):
            for j in range(3):
                for k in range(3):
                    Y[i*3+j,j*3+k] = mag_conf[mag_is[i]][k]
                #Y[i*3+j,9] = mag_conf[sym_type(mag_is[i]+1,sym)-1][j]
                Y[i*3+j,9] = mag_conf[sym.permutations[mag_is[i]+1]-1][j]

        if debug: 
            print('')
            print('Matrix of the system to be solved.')
            sympy.pprint(Y)
        
        Rs = matrix('s',3,name='R') 
        sol = sympy.solve_linear_system(Y,Rs.x[0,0],Rs.x[0,1],Rs.x[0,2],Rs.x[1,0],Rs.x[1,1],Rs.x[1,2],Rs.x[2,0],\
                                         Rs.x[2,1],Rs.x[2,2])
        if debug:
            print('solution of the system')
            print(sol)

        for s in sol:
            Rs.subs(s,sol[s])

        if debug:
            print('')
            print('general form of transformation matrix that keeps the magnetic order invariant:')
            sympy.pprint(Rs.mat())
            print('')

        for mat in mats:
            if debug:
                print('Considering the matrix:')
                sympy.pprint(mat)
            fits = True
            for i in range(3):
                for j in range(3):
                    if Rs.x[i,j] in sol:
                        if mat[i,j] != sol[Rs.x[i,j]]:
                            fits = False

            if debug:
                if fits:
                    print('The matrix is compatible with the magnetic order.')
                else:
                    print('The matrix is not compatible with the magnetic order.')

            if fits:
                if mat.det() == 1 and sym.has_T:
                    fits = False
                    if debug:
                        print('Improper spin rotation, thus not taking this matrix.')
            if fits:
                sym_new = sym.copy()
                sym_new.Rs = mat
                syms_noso.append(sym_new)

        if debug:
            print('original symmetry operation:')
            print('space part:')
            sympy.pprint(sym.R)
            print('magnetic part:')
            sympy.pprint(sym.Rs)
            print('')

            print('new symmetry operations (magnetic part only):')
            for sm in syms_noso[start_new:]:
                sympy.pprint(sm.Rs)
                print('')

    return syms_noso




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Module for parsing user input."""
from __future__ import absolute_import
from __future__ import division

from builtins import range
from past.utils import old_div
from builtins import object
import argparse
import sys
import textwrap
from math import log10
from .version import __version__
import sys

class InputError(Exception):
    pass

class options(object):
    """Class to store all the input options.

    Is meant to be used only in conjuction with the parse function.
    Initialize by running
        opt = options(args),
    where args is a dictionary of all arguments.
    Check method controls whether there are comflicting options, however this is not complete and it does not
    check whether the options have allowed values!
    You can access the options by:
        opt['option']
    and also change by:
        opt['option'] = new_value
    """
    def __init__(self,args):
        """During init only the mandatory arguments are set, the rest are set to default values."""
        self.args = args
        self.check()

    def __setitem__(self,key,value):
        self.args[key] = value
        self.check()

    def __getitem__(self,key):
        return self.args[key]

    def __str__(self):
        return self.args.__str__()

    def check(self):
        if self['transform_result'] and self['transform_syms']:
            raise InputError('You cannot specify both --transform-result and --transform-syms')

        if self['mode'] == 'res':
            if self['atom2'] != -1:
                if self['atom'] == -1:
                    raise InputError('projection2 can be setonly if projection1 is set')

            if self['inp'] and self['group']:
                raise InputError('You cannot specify both the symmetry group and Findsym input file')

            if (not self['inp']) and (not self['group']):
                raise InputError('You have to specify either the symmetry group or the Findsym input file')

            if ( self['atom'] != -1 or self['atom2'] !=-1 ) and self['group']:
                raise InputError('Projections not possible with group name input. Use Findsym input instead.')

            if self['equiv'] and self['group']:
                raise InputError('Equivalent configurations are not possible with group name input. Use findsym input file.')

            if self['noso'] and self['group']:
                raise InputError('Spin-orbit coupling cannot be ignored when group name is used as an input.')

            if self['noso'] and (self['equiv'] or (self['exp'] != -1) or (self['atom2'] != -1)):
                raise InputError('This is not implemented.')

            if (self['equiv'] or (self['exp'] != -1)) and self['syms_sel'] != -1:
                raise InputError('You cannot select symmetries when exp or equiv is set since this is not implemented')

            if self['equiv'] and ( self['exp'] != -1 ):
                raise InputError('You cannot use --equiv and --exp together. Equivalent configurations not supported\
                        for expansions.')
        if self['mode'] == 'mham':
            if self['group'] is not None:
                raise InputError('group input is not allowed for mham')
            if self['inp'] is None:
                raise InputError('You need to specify findsym input file')
        if not self['inp'] and self['print_pos']:
            raise InputError('print-pos is only possible with findsym input.')

def parse(clargs=None):
    """Parses the input line arguments

    If clargs is not set then this parses the input arguments. Otherwise clargs is a string that contains the input arguments
    like on command line.
    Args:
        clargs (optional[string]): The input arguments to be parsed.
    """

    parser_parent = argparse.ArgumentParser(add_help=False)
    parser_parent.add_argument('-f','--findsym',help='Findsym input file',default=None,dest='inp')
    parser_parent.add_argument('-b','--basis',help='Sets a coordinate basis: abc for conventional crystallographic basis, i for the one used in input \
    (default). cart for a cartesian basis in which the input basis is define. \
    abc_c for orthogonalized crystalographic basis (not tested much).',default='cart')
    parser_parent.add_argument('--print-syms',action='store_const',const=True,default=False,help='Prints all symmetry operations.')
    parser_parent.add_argument('--transform-result',action='store_const',const=True,default=False,help=
            'Keep the symmetry operations in findsym basis and transform the result to the correct basis.')
    parser_parent.add_argument('--transform-syms',action='store_const',const=True,default=False,help=\
            'Transform the symmetry operations to the correct basis instead of transforming the result.')
    parser_parent.add_argument('--syms',default=-1,help='Choose which symmetry operations to take, the rest is ignored.\
            Insert symmetry operation numbers separated by commas with no spaces. They are numbered as they appear\
            in the findsym output file. Also can include ranges. Example: 1-3,7,9-12',dest='syms_sel')
    parser_parent.add_argument('--print-opt',action='store_const',const=True,default=False,help='If set all the input options are printed.')
    parser_parent.add_argument('-g','--group',help='group name',default=None)
    parser_parent.add_argument('--latex',action='store_const',const=True,default=False,help='If set, the matrices are printed also in a latex format.')
    parser_parent.add_argument('--print-pos',action='store_const',const=True,default=False,help='If set prints the atomic sites used in\
            findsym.')
    parser_parent.add_argument('--symbolic',action='store_true',default=False,help=
    'If chosen, the symbolical mode will be used. This means that the symmetrization is done exactly without taking \
     into account finie precision of the input. Use carefully as it can fail in some cases !!!')
    parser_parent.add_argument('--num-prec',dest='num_prec',default=1e-3,help='The numerical precision.\
            Within the alghoritm for symmetrizing, two numbers are considered the same when they are closer\
            then num-prec. Default value 1e-3. Setting this the same as precision for findsym should be a safe choice.' )
    parser_parent.add_argument('--round-prec',dest='round_prec',default=None,help='In the numerical mode, this gives the \
            number of digits for rounding. Default value is int(log10(1/num_prec))')
    parser_parent.add_argument('--dont-round',action='store_const',const=True,default=False,help=
            'If selected, no rounding is done.')
    parser_parent.add_argument('--pos-prec',dest='pos_prec',default=None,help='Precision for the positions. \
    This refers to the positions written in the basis of the unic cell vectors. \
    The positions are read from the findsym output.\
    Default value is maximum of 1.1e-5 and findsym precision (second line of the findsym input file.')
    parser_parent.add_argument('--print-format',dest='print_format',default=None,type=int,help='Format for printing')
    parser_parent.add_argument('--version',action='store_const',const=True,default=False,help=
            'Print version.')

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,\
            description=textwrap.dedent('''\
            There are two modes: res and mham.
                * res is for symmetry of response functions                *
                * mham is for symmetry of magnetic Hamiltonians
            Use symmetr [res,mham] -h for description of input arguments of each mode. '''))
    subparsers = parser.add_subparsers(dest='mode')
    parser_res = subparsers.add_parser('res',parents=[parser_parent])
    parser_mham = subparsers.add_parser('mham',parents=[parser_parent])

    parser_res.add_argument('op1', help='The opeator of which response we are evaluating.',metavar='A')
    parser_res.add_argument('op2', help='The field which induces the response.',metavar='F')
    parser_res.add_argument('-p','--projection',help='Sets a projection on an atom.',default=-1,dest='atom',type=int)
    parser_res.add_argument('-p2','--projection2',help='Sets a projection on a second atom. Tries to find a relation between tensors on the first \
            atom and on the second atom.',default=-1,dest='atom2',type=int)
    parser_res.add_argument('-e','--equivalent',action='store_const',const=True,default=False,help='finds response matrices for equivalent magnetic configurations.',dest='equiv')
    parser_res.add_argument('--no-rename',action='store_true')
    parser_res.add_argument('--debug',help='Controls if debug output is printed. all means all debug output is printed, symmetrize means debug\
            output for symmetrizing, rename for renaming, equiv for finding the equivalent configurations,\
            noso for the symmetry without spin-orbit coupling. op1eqop2 is also possible.',default='')
    parser_res.add_argument('--exp',default=-1,help=\
            'Prints the tensor, which describes the expansion of the linear response tensor in magnetic moments.\
            Choose which order term in the expansion.\
            Only works for ferromagnets and collinear antiferromagnets. In antiferromagnets only for quantities\
            on a sublattice!!!',type=int)
    parser_res.add_argument('--syms-noso',default=-1,help='Like --syms, but fot noso symmetry operations',dest='syms_sel_noso')
    parser_res.add_argument('--noso',action='store_const',const=True,default=False,help=
            'Symmetry without spin-orbit coupling. Very experimental.')
    parser_res.add_argument('--same-op-sym',dest='same_op_sym',action='store_true',default=None)
    parser_res.add_argument('--ignore-same-op-sym',dest='same_op_sym',action='store_false',default=None)
    parser_res.add_argument('--sym-inds',default=None,help='')
    parser_res.add_argument('--asym-inds',default=None,help='')

    parser_mham.add_argument('-s','--sites',help='Atomic sites for which the Magnetic Hamiltonian is considered.\
            List of integeres separated by commas with no spaces, e.g. 1,2. Corresponds to the order of the \
            Hamiltonian. Number of sites must be even!',required=True)
    parser_mham.add_argument('--debug',help='Controls if debug output is printed.',default='')
    parser_mham.add_argument('-e','--equivalent',action='store_const',const=True,default=False,\
            help='Finds the magnetic Hamiltonian also magnetic sites related to the input one by a symmetry operation.',\
            dest='equiv')

    if clargs != None:
        args = parser.parse_args(clargs.split())
    else:
        args = parser.parse_args()

    args_dict = vars(args)


    args_dict['debug'] = args_dict['debug'].split(',')

    #set the debug variables
    args_dict['debug_sym'] = False
    args_dict['debug_rename'] = False
    args_dict['debug_equiv'] = False
    args_dict['debug_tensor'] = False
    args_dict['debug_time'] = False
    args_dict['debug_symY'] = False
    args_dict['debug_noso'] = False
    args_dict['debug_op1eqop2'] = False

    if 'symmetrize' in args_dict['debug'] or 'all' in args_dict['debug'] or 'symmetrizeY' in args_dict['debug']:
        args_dict['debug_sym'] = True
    if 'rename' in args_dict['debug'] or 'all' in args_dict['debug']:
        args_dict['debug_rename'] = True
    if 'equiv' in args_dict['debug'] or 'all' in args_dict['debug']:
        args_dict['debug_equiv'] = True
    if 'exp' in args_dict['debug'] or 'all' in args_dict['debug']:
        args_dict['debug_tensor'] = True
    if 'time' in args_dict['debug'] or 'all' in args_dict['debug']:
        args_dict['debug_time'] = True
    if 'symmetrizeY' in args_dict['debug']:
        args_dict['debug_symY'] = True
    if 'noso' in args_dict['debug']:
        args_dict['debug_noso'] = True
    if 'op1eqop2' in args_dict['debug']:
        args_dict['debug_op1eqop2'] = True

    if args_dict['mode'] == 'mham':
        args_dict['sites'] = args_dict['sites'].split(',')
        for i,x in enumerate(args_dict['sites']):
            args_dict['sites'][i] = int(x)

    #this sets the op variables
    def convert_op1(op1):
        if op1 == 'j':
            return 'v'
        elif op1 == 'B':
            return 's'
        else:
            return op1
    def convert_op2(op2):
        if op2 == 'E':
            return 'v'
        else:
            return op2

    if args_dict['mode'] == 'res':
        op_types1 = args_dict['op1'].split('.')
        for i,op in enumerate(op_types1):
            op_types1[i] = convert_op1(op)
        if args_dict['op2'] == '0':
            op_types2 = []
        else:
            op_types2 = args_dict['op2'].split('.')
            for i,op in enumerate(op_types2):
                op_types2[i] = convert_op2(op)
        args_dict['op_types'] = op_types1 + op_types2
        args_dict['op_lengths'] = (len(op_types1),len(op_types2))
        args_dict['op3'] = None

    if args_dict['mode'] == 'res':
        if args_dict['same_op_sym'] is None:
            if len(args_dict['op_types']) == 2:
                args_dict['same_op_sym'] = True
            else:
                args_dict['same_op_sym'] = False

    if args_dict['symbolic']:
        args_dict['num_prec'] = None
    else:
        args_dict['num_prec'] = float(args_dict['num_prec'])

    if not args_dict['symbolic']:
        if args_dict['round_prec'] is None:
            args_dict['round_prec'] = int(log10(old_div(1,args_dict['num_prec'])))
        else:
            args_dict['round_prec'] = int(args_dict['round_prec'])
        if args_dict['dont_round']:
            args_dict['round_prec'] = None
    
    if args_dict['pos_prec'] is not None:
        args_dict['pos_prec'] = float(args_dict['pos_prec'])

    def parse_sym_inds(sym_inds):
        if sym_inds is not None:
            sym_inds_c = args_dict['sym_inds'].split('::')
            sym_inds_c[0] = sym_inds_c[0].split(':')
            sym_inds_c[1] = sym_inds_c[1].split(':')
            sym_inds = [[],[]]
            for i in range(2):
                for ind in sym_inds_c[i]:
                    if ind != '':
                        (j,Pj) = ind.split(',')
                        sym_inds[i].append((int(j),int(Pj)))

        return sym_inds

    if args_dict['mode'] == 'res':
        sym_inds = parse_sym_inds(args_dict['sym_inds'])
        asym_inds = parse_sym_inds(args_dict['asym_inds'])

    opt = options(args_dict)

    if (not opt['transform_result']) and (not opt['transform_syms']):
            opt['transform_syms'] = True

    return opt



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Contains functions for obtaining list of symmetry operations and transformation matrix
based on user input.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from builtins import range
from past.utils import old_div
from copy import deepcopy

from . import fslib
import sympy
from sympy import sympify as spf
from .groups import group_sym
from .noso import noso_syms
from .symmetry import findsym2sym,matsym2sym
from fractions import Fraction

def make_rational(mat):
    """Converts sympy matrix to a rational number form.

    This is useful since sympy can work exactly with rational numbers so there are no floating point accuracy issues.
    """

    ncols = mat.cols
    nrows = mat.rows
    mat_r = sympy.zeros(mat.rows,mat.cols)

    for i in range(nrows):
        for j in range(ncols):
            if type(mat[i,j]) == str and '\\' in mat[i,j]:
                mat_r[i,j] = sympy.sympify(Fraction(mat[i,j]))
            else:
                mat_r[i,j] = sympy.sympify(Fraction(float(mat[i,j])))

    return mat_r

def create_Tm(vec_a,vec_b,vec_c):

    T = sympy.zeros(3)

    T[0,0] = vec_a[0]
    T[1,0] = vec_a[1]
    T[2,0] = vec_a[2]

    T[0,1] = vec_b[0]
    T[1,1] = vec_b[1]
    T[2,1] = vec_b[2]

    T[0,2] = vec_c[0]
    T[1,2] = vec_c[1]
    T[2,2] = vec_c[2]

    T = make_rational(T)

    return T

def create_Ti(fin):

    T_i = sympy.zeros(3)

    inp_type = int(fin[2])

    if inp_type == 1:

        vec_1 = fin[3].split()
        vec_2 = fin[4].split()
        vec_3 = fin[5].split()

        T_i[0,0] =  spf(vec_1[0])
        T_i[1,0] =  spf(vec_1[1])
        T_i[2,0] =  spf(vec_1[2])

        T_i[0,1] =  spf(vec_2[0])
        T_i[1,1] =  spf(vec_2[1])
        T_i[2,1] =  spf(vec_2[2])

        T_i[0,2] =  spf(vec_3[0])
        T_i[1,2] =  spf(vec_3[1])
        T_i[2,2] =  spf(vec_3[2])

    if inp_type == 2:

        a,b,c,al,bet,gam = fin[3].split()
        al = al+'*2*pi/360'
        bet = bet+'*2*pi/360'
        gam = gam+'*2*pi/360'
        gam2 = 'acos((cos({gam})-cos({bet})*cos({al}))/(sin({al})*sin({al})))'.format(gam=gam,al=al,bet=bet)        

        
        T_i[0,0] =  spf('{a}*sin({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
        T_i[1,0] =  spf('{a}*cos({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
        T_i[2,0] =  spf('{a}*cos({bet})'.format(a=a,gam2=gam2,bet=bet))

        T_i[0,1] =  spf(0)
        T_i[1,1] =  spf('{b}*sin({al})'.format(b=b,al=al))
        T_i[2,1] =  spf('{b}*cos({al})'.format(b=b,al=al))

        T_i[0,2] =  spf(0)
        T_i[1,2] =  spf(0)
        T_i[2,2] =  spf(c)
    
    return T_i

def is_hex(lines):
    for i,line in enumerate(lines):
        if 'Values of a,b,c,alpha,beta,gamma:' in line:
            pos = i
    angles = lines[pos+1].split()[3:6]
    hexag= False
    for angle in angles:
        if int(round(float(angle))) != 90:
            if int(round(float(angle))) == 120:
                hexag = True
            else:
                sys.exit('one of the angles in findsym output is neither 90 nor 120.')
    return hexag

def get_syms(opt):
    if opt['inp']:
        #runs findsym and reads the output
        lines = fslib.run_fs(opt['inp'])
        syms = fslib.r_sym(lines,num_prec = opt['pos_prec'])

    if opt['group']:
        atom = -1
        _,syms=group_sym(opt['group'],debug=False)

    #this selects some of the symmetries 
    if opt['syms_sel'] != -1:
        syms_sel = opt['syms_sel'].split(',')
        syms_sel2 = []
        for i in range(len(syms_sel)):
            if '-' in syms_sel[i]:
                s = syms_sel[i].split('-')
                syms_sel2 += list(range(int(s[0]),int(s[1])+1))
            else:
                syms_sel2.append(int(syms_sel[i]))

        syms_new = []
        for i in range(len(syms)):
            if i+1 in syms_sel2:
                syms_new.append(syms[i])

        syms = syms_new

    syms_g = []
    for sym in syms:
        syms_g.append(findsym2sym(sym))
    syms = syms_g

    return syms

def get_T(opt,nonmag=False):
    """
    Returns transformation matrix from the conventional coordinate system (used by findsym)
    to the user selected basis.

    There are two different behaviors:
        nonmag=False: conventional coordinate system for the magnetic system
        nonmag=True: conventional coordinate system for the nonmagnetic system
    """
    if opt['inp'] is not None:
        #runs findsym and reads the output
        fin = fslib.read_fs_inp(opt['inp'],clean=False)
        lines = fslib.run_fs(opt['inp'])
        if nonmag:
            lines_nm = fslib.run_fs_nonmag(opt['inp'])

        #reads the input
        #vec_a,b,c are needed to know the basis transformation
        #syms contain the symmetries in the form that is needed by symmetr
        if not nonmag:
            [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
        else:
            [vec_a,vec_b,vec_c] = fslib.r_basis(lines_nm)

        #construct the transformation matrix from the findsym basis to the selected basis
        if 'abc' == opt['basis']:

            T = sympy.Matrix(sympy.Identity(3))

        if 'i' == opt['basis']:

            T = create_Tm(vec_a,vec_b,vec_c)

        if 'cart' == opt['basis']:

            T_m = create_Tm(vec_a,vec_b,vec_c)

            T_i = create_Ti(fin)

            T = T_i*T_m

        if 'custom' == opt['basis']:
            
            #T_i is a transformation matrix from the input coordinate system to the cartesian one
            T_i = create_Ti(fin)
                    
            for i in range(len(fin)):
                if 'axes:' in fin[i]:
                    loc = i
            
            vec_1 = fin[loc+1].split()
            vec_2 = fin[loc+2].split()
            vec_3 = fin[loc+3].split()

            #T_c is the transformation matrix from the used-defined basis to the cartesian one
            T_c = sympy.zeros(3)

            T_c[0,0] =  spf(vec_1[0])
            T_c[1,0] =  spf(vec_1[1])
            T_c[2,0] =  spf(vec_1[2])

            T_c[0,1] =  spf(vec_2[0])
            T_c[1,1] =  spf(vec_2[1])
            T_c[2,1] =  spf(vec_2[2])

            T_c[0,2] =  spf(vec_3[0])
            T_c[1,2] =  spf(vec_3[1])
            T_c[2,2] =  spf(vec_3[2])

            normalize = True
            if normalize == True:
                norm = sympy.sqrt(T_c[0,0]**2 + T_c[1,0]**2 + T_c[2,0]**2)
                T_c[0,0] = old_div(T_c[0,0], norm)
                T_c[1,0] = old_div(T_c[1,0], norm)
                T_c[2,0] = old_div(T_c[2,0], norm)

                norm = sympy.sqrt(T_c[0,1]**2 + T_c[1,1]**2 + T_c[2,1]**2)
                T_c[0,1] = old_div(T_c[0,1], norm)
                T_c[1,1] = old_div(T_c[1,1], norm)
                T_c[2,1] = old_div(T_c[2,1], norm)

                norm = sympy.sqrt(T_c[0,2]**2 + T_c[1,2]**2 + T_c[2,2]**2)
                T_c[0,2] = old_div(T_c[0,2], norm)
                T_c[1,2] = old_div(T_c[1,2], norm)
                T_c[2,2] = old_div(T_c[2,2], norm)

            T_m = create_Tm(vec_a,vec_b,vec_c)

            T = T_c.inv()*T_i*T_m

        if 'abc_c' == opt['basis']:

            T = sympy.zeros(3)

            if not nonmag:
                abc = fslib.r_abc(lines)
            else:
                abc = fslib.r_abc(lines_nm)

            a = abc[0]
            b = abc[1]
            c = abc[2]
            al = abc[3]
            bet = abc[4]
            gam = abc[5]

            al = al+'*2*pi/360'
            bet = bet+'*2*pi/360'
            gam = gam+'*2*pi/360'
            gam2 = 'acos((cos({gam})-cos({bet})*cos({al}))/(sin({al})*sin({al})))'.format(gam=gam,al=al,bet=bet)        
            
            T[0,0] =  spf('{a}*sin({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
            T[1,0] =  spf('{a}*cos({gam2})*sin({bet})'.format(a=a,gam2=gam2,bet=bet))
            T[2,0] =  spf('{a}*cos({bet})'.format(a=a,gam2=gam2,bet=bet))

            T[0,1] =  spf(0)
            T[1,1] =  spf('{b}*sin({al})'.format(b=b,al=al))
            T[2,1] =  spf('{b}*cos({al})'.format(b=b,al=al))

            T[0,2] =  spf(0)
            T[1,2] =  spf(0)
            T[2,2] =  spf(c)

    if opt['group'] is not None:
        atom = -1
        hex_group,_=group_sym(opt['group'],debug=False)

        if 'i' == opt['basis'] or 'abc' == opt['basis']:
            print('Using the conventional coordinate system!')
            T = sympy.Matrix(sympy.Identity(3))

        if 'cart' == opt['basis']:

            print('Using a cartesian coordinate system')
            if hex_group:
                T = sympy.zeros(3)
                T[0,0] = 1
                T[0,1] = sympy.sympify(Fraction(-0.5))
                T[0,2] = 0
                T[1,0] = 0 
                T[1,1] = old_div(sympy.sqrt(3),2)
                T[1,2] = 0
                T[2,0] = 0
                T[2,1] = 0
                T[2,2] = 1
            else:
                T = sympy.Matrix(sympy.Identity(3))

    return T

def get_syms_nonmag(opt):
    lines = fslib.run_fs_nonmag(opt['inp'])
    syms = fslib.r_sym(lines)
    
    syms_g = []
    for sym in syms:
        syms_g.append(findsym2sym(sym))
    syms = syms_g

    return syms

def get_syms_noso(opt):

    fin_c = fslib.read_fs_inp(opt['inp'])
    mags = fslib.r_mag_fin(fin_c)
    lines = fslib.run_fs(opt['inp'])
    lines_nm = fslib.run_fs_nonmag(opt['inp'])
    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    [vec_a_nm,vec_b_nm,vec_c_nm] = fslib.r_basis(lines_nm)

    Tm = create_Tm(vec_a,vec_b,vec_c)
    Tnm = create_Tm(vec_a_nm,vec_b_nm,vec_c_nm)
    mags_T = convert_vecs(mags,Tm.inv())
    syms_nm = get_syms_nonmag(opt)
    for sym in syms_nm:
        sym.convert(Tm.inv()*Tnm,in_place=True)
    hexag = is_hex(lines)
    syms_noso = noso_syms(syms_nm,mags_T,hexag,debug=opt['debug_noso'])

    if opt['syms_sel_noso'] != -1:
        syms_sel_noso = opt['syms_sel_noso'].split(',')
        syms_sel_noso2 = []
        for i in range(len(syms_sel_noso)):
            if '-' in syms_sel_noso[i]:
                s = syms_sel_noso[i].split('-')
                syms_sel_noso2 += list(range(int(s[0]),int(s[1])+1))
            else:
                syms_sel_noso2.append(int(syms_sel_noso[i]))

        syms_noso_new = []
        for i in range(len(syms_noso)):
            if i+1 in syms_sel_noso2:
                syms_noso_new.append(syms_noso[i])

        syms_noso = syms_noso_new

    return syms_noso

def get_mags(inp):

    fin_c = fslib.read_fs_inp(inp)
    mags = fslib.r_mag_fin(fin_c)

    return mags

def convert_vec(vec,T):
    vec_T = T*vec.T 
    vec_T = vec_T.T
    return vec_T

def convert_vecs(mags,T):
    """Converts a list of vectors by transform matrix T.
    """

    mags_T = []
    for i,mag in enumerate(mags): 
        mag_T = convert_vec(mag,T)
        mags_T.append(mag_T)

    return mags_T

def get_Tm(inp,nonmag=False):

    if not nonmag:
        lines = fslib.run_fs(inp)
    else:
        lines = fslib.run_fs_nonmag(inp)

    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    Tm = create_Tm(vec_a,vec_b,vec_c)

    return Tm

def get_Tm_fin(inp_list,nonmag=False):
    if not nonmag:
        lines = fslib.run_fs_fin(inp_list)
    else:
        lines = fslib.run_fs_fin(fslib.make_fsinp_nonmag(inp_list))

    [vec_a,vec_b,vec_c] = fslib.r_basis(lines)
    Tm = create_Tm(vec_a,vec_b,vec_c)

    return Tm

def get_metric(opt,debug=False):

    if opt['group'] is not None:
        #the metric tensor is trivial in the cartesian coordinate system
        G = sympy.Matrix([[1,0,0],[0,1,0],[0,0,1]])
        opt2 = deepcopy(opt)
        T = get_T(opt2)
        G = T.inv() * G * T.inv().T

    else:
        if debug:
            print('')
            print('Get_metric debug output')
            print('================================')
        #read the basis vectors used by findsym
        #they are given in the basis used in the findsym input
        lines = fslib.run_fs(opt['inp'])
        [vec_a,vec_b,vec_c] = fslib.r_basis(lines)

        vec_a = sympy.Matrix(vec_a)
        vec_b = sympy.Matrix(vec_b)
        vec_c = sympy.Matrix(vec_c)

        #create transformation matrix from the input basis to a cartesian basis
        fin = fslib.read_fs_inp(opt['inp'],clean=False)
        T_i = create_Ti(fin)

        #Transform to a cartesian coordinate system
        vec_a = T_i * vec_a
        vec_b = T_i * vec_b
        vec_c = T_i * vec_c

        #calculate the dual basis vectors
        V = vec_a.dot(vec_b.cross(vec_c))
        vec_A = old_div(vec_b.cross(vec_c),V)
        vec_B = old_div(vec_c.cross(vec_a),V)
        vec_C = old_div(vec_a.cross(vec_b),V)

        if debug:

            print('basis vectors')
            print(vec_a,vec_b,vec_c)
            print('dual basis vectors')
            print(vec_A,vec_B,vec_C)

        A_l = vec_a.row_join(vec_b)    
        A_l = A_l.row_join(vec_c)

        A_L = vec_A.row_join(vec_B)    
        A_L = A_L.row_join(vec_C)

        G = A_l.LUsolve(A_L)
        if debug:
            print('The metrix tensor')
            sympy.pprint(G)
            print('The inverse of the metrix tensor')
            sympy.pprint(G.inv())

        if debug:
            print('metric tensor test')
            for i in range(3):
                a_I = 0*vec_a
                for j in range(3):
                    a_I += G.T[i,j]*A_l[:,j]
                print(a_I)

            print('Get metric debug output end')
            print('================================')
            print('')

    return G




from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import range
from builtins import object
import re
import copy
import sys
import math
import time

import sympy
import numpy as np
import numpy.linalg
import mpmath   

from .tensors import matrix, mat2ten
from .fslib import transform_position
from .conv_index import *

class params_trans(object):
    def __init__(self,op1,op2,op3,l,T=None,sym_format='findsym'):
        self.op1 = op1
        self.op2 = op2
        self.op3 = op3
        self.l = l
        self.T = T
        self.sym_format = sym_format

class SymmetrOpt(object):
    def __init__(self,num_prec=None,debug=False,debug_time=False,debug_Y=False,round_prec=None):
        self.num_prec = num_prec
        self.debug = debug
        self.debug_time = debug_time
        self.debug_Y = debug_Y
        self.round_prec = round_prec

def symmetr(syms,X,trans_func,params,opt=None):
    """
    This symmetrizes a tensor X given a list of symmetries and a transformation function.

    This function should be quite general and is now used for all symmetrizing.

    Args:
        syms: list of symmetry operations
        X: tensor - must be a tensor class
        trans_func: function that transforms the tensor X using symmetry sym
            trans_func must work in the following way:
            X_trans = trans_func(X,sym,params)
            If trans_func returns None then the symmetry operation is ignored
        params: parameters to be sent to function trans_func

    Returns:
        X_trans: the symmetry restricted form of tensor X
    """
    if opt is None:
        opt = SymmetrOpt()
    num_prec = opt.num_prec
    debug = opt.debug
    debug_time = opt.debug_time
    debug_Y = opt.debug_Y

    if debug:
        print('')
        print('======= Starting symmetrizing =======')

    #we do a loop over all symmetry operations, for each symmetry, we find what form the response matrix can have, when the system has this symmetry
    #for next symmetry we take the symmetrized matrix from the previous symmetry as a starting point
    for sym in syms:
        
        if debug:
            print('Symmetry:') 
            print(sym)
            print('')

        X_trans = trans_func(X,sym,params)
        if X_trans is None:
            continue
            
        if debug:
            print('')
            print('Current form of the tensor:')
            print('')
            X.pprint()
            print('')
            print('Transformed tensor:')
            print('')
            X_trans.pprint()
        #X.pprint()
        #X_trans.pprint()
        #The tensor must be equal to the transformed tensor, this give us a system of linear equations.
        #matrix Y is a matrix that represents this system, ie the system X-X_trans = 0
        #we reverse the order of the rows
        # it doesn't really matter but the results are more natural this way

        if debug_time:
            t1 = time.clock()

        Y = sympy.zeros(X.dim1**X.dim2)

        rev_inds = list(reversed(X.inds))
        #we do a loop over all rows of the matrix Y - ie over all linear equations
        n = -1
        for ind1 in X:
            n += 1
            m = -1

            #if this is zero, then we do not have to do any substituting so this saves quite a lot of time
            if X[ind1]-X_trans[ind1] == 0:
                is_zero = True
            else:
                is_zero = False

            speed_test = True
            if speed_test:
                inds = re.findall(r'x[0-9]+',sympy.srepr(X[ind1]-X_trans[ind1]))
                for ind2 in reversed(inds):
                    m_index = (int(i) for i in re.findall(r'[0-9]',ind2))
                    m = rev_inds.index(tuple(m_index))
                    Y_p = X[ind1]-X_trans[ind1]
                    #now in the equation we substite 1 to the matrix component that correponds to the column and 0 to all others
                    for ind3 in inds:
                        if ind2 == ind3:
                            Y_p = Y_p.subs(X[ind3],1)
                        else:
                            Y_p = Y_p.subs(X[ind3],0)

                    Y[n,m] = Y_p

            else:
                for ind2 in rev_inds:
                    m += 1
                    Y_p = X[ind1]-X_trans[ind1]
                    #now in the equation we substite 1 to the matrix component that correponds to the column and 0 to all others
                    for ind3 in X.inds:
                        if ind2 == ind3:
                            Y_p = Y_p.subs(X.x[ind3],1)
                        else:
                            Y_p = Y_p.subs(X.x[ind3],0)

                    Y[n,m] = Y_p

        if debug_time:
            t2 = time.clock()
            print('Time for constructing Y: ', t2-t1)
            t1 = time.clock()

        #this transforms the matrix into the Reduced row echelon form
        #piv are the indeces o the pivot columns
        if num_prec is None:
            [rref,piv] = Y.rref()
        else:
            def zerofunc(x):
                try:
                    a = abs(x) < num_prec
                    if a:
                        result = True
                    else:
                        result = False
                except:
                    result = None
                #print(result)
                return result
            #sympy.pprint(Y)
            #[rref,piv] = Y.rref()        
            [rref,piv] = Y.rref(iszerofunc=lambda x:abs(x)<num_prec)        
            #[rref,piv] = Y.rref(iszerofunc=zerofunc)        

        if debug_time:
            t2 = time.clock()
            print('Time for reducing Y to reduced row echelon form: ', t2-t1)

        if debug_Y:
            print('')
            print('Matrix representing the linear equation system that has to be satisfied: (right hand side is zero)')
            sympy.pprint(Y)
            print('')
            print('Reduced row echelon form and indeces of the pivot columns:')
            sympy.pprint(rref)
            print(piv)
            print('')

        #a loop over all the pivots: it's the pivots that give interesting information
        for j in list(reversed(piv)):

            
            #find the row of pivot j
            found = False
            i = X.dim1**X.dim2-1
            while found == False:
                if rref[i,j] == 1:
                    found = True
                else:
                    i = i-1
            
            if debug:
                print('')
                print('considering pivot ', i,j)

            tmp = 0
            #now we just make use of the linear equation that holds for this pivot
            #keep in mind that the rows are in reversed order
            for ll in range(j+1,X.dim1**X.dim2):
                tmp = tmp - rref[i,ll]*X.x[rev_inds[ll]]
            X = X.subs(X.x[rev_inds[j]],tmp)

            if debug:
                print('substituting ', end=' ')
                sympy.pprint(X.x[rev_inds[j]])
                print(' for ', end=' ')
                sympy.pprint(tmp)
                print('')

        if debug:
            print('Current form of the tensor:')
            X.pprint()
            print('')

    if opt.round_prec is not None:
        X.round(opt.round_prec)
        
    if debug:
        print('Symmetrized tensor:')
        X.pprint()
        print('')
        print('======= End symmetrizing =======')

    return X

def even_odd(Xs):
    """Finds whether the first part of the response tensor is even or odd
    
    Args:
        op1,op2,op3: the operator types
        Returns: either ('even','odd') or ('odd','even')
    """
    if Xs[0].is_even() and not Xs[1].is_even():
        return('even','odd')
    elif not Xs[0].is_even() and Xs[1].is_even():
        return('odd','even')
    else:
        raise Exception('Wrong transformation under time-reversal')


def symmetrize_res(symmetries,X,proj=-1,s_opt=None):

    syms_sel = []
    for sym in symmetries:
        
        if s_opt.debug:
            print('Symmetry:') 
            print(sym)
            print('')
            if proj != -1:
                print('Symmetry transforms the atom ', proj, ' into atom ', sym.permutations[proj])
                if  sym.permutations[proj] != proj:
                    print('Skipping symmetry')
                    print('')

        #if there is a projection set up we only consider symmetries that keep the atom invariant
        if proj == -1 :
            take_sym = True
        elif sym.permutations[proj] == proj:
            take_sym = True
        else:
            take_sym = False

        if take_sym:
            syms_sel.append(sym)

    def trans_func(X,sym,params):
        return X.transform(sym)
    X = symmetr(syms_sel,X,trans_func,None,s_opt)

    return X

def symmetrize_same_op(X,s_opt=None):

    perm = {}
    perm[0] = 1
    perm[1] = 0
    perms=[perm] 

    def trans_func(X,perm,params):
        X_T = X.copy0()

        for ind in X:
            ind_T = [0]*len(ind)
            for i in range(len(ind)):
                ind_T[i] = ind[perm[i]]
            ind_T = tuple(ind_T)
            X_T[ind_T] = X.T_comp * X[ind]

        ind_types = [0] * X.dim2
        for i in range(X.dim2):
            ind_types[i] = X.ind_types[perm[i]]
        X_T.ind_types = tuple(ind_types)

        for i in range(X.dim2):
            if X_T.ind_types[i] != X.ind_types[i]:
                X_T.reverse_index(i)

        return X_T

    X = symmetr(perms,X,trans_func,None,s_opt)
        
    return X

def symmetr_AB(syms,X,atom1,atom2,round_prec=None):
    """
    Tries to transform the tensor projected on one atom to a different atom

    Args:
        syms: The symmmetry operations. Format as outputted by read.py
        X: The input tensor.
        op1: The first operator.
        op2: The second operator.
        atom1: The atom on which X is projected.
        atom2: The atom on which X is transformed.
        T (Optional[matrix]): Coordinate transformation matrix. If it is set, the symmetry operations will be transformed by this matrix.
            Symmetry operations are given in basis A. T transforms from A to B, ie Tx_A = x_B.

    Returns:
        X_trans: The transformed tensor.
    """

    X_trans = []

    found = False
    for sym in syms:
        #there will usually be more symmetries that transform from atom1 to atom2, we need only one, as they all
        #give the same results
        if sym.permutations[atom1] == atom2 and not found:
            found = True
            for l in range(2):
                X_trans.append(X[l].transform(sym))

    if found:
        if round_prec is not None:
            for X in X_trans:
                X.round(round_prec)
        return X_trans
    else:
        return None



from __future__ import print_function
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from builtins import range
from builtins import object

from collections import OrderedDict
import sympy

class confs(object):
    """
    Class to store equivalent magnetic configurations and their response tensors.

    Description:
        self.nconfs is the number of configurations stored
        self.confs[i] is the configuration (stored as a dict) with number i
        self.Xs[i] is the tensor of the configuration i

    Usage:
        __init__ creates an empty object. 
        Add a configuration by add(conf,X).
        Print all by pprint.
    """

    def __init__(self):
        """Creates an empty confs object."""

        self.confs = {}
        self.Xs = {}
        self.nconfs = 0

    def add(self,conf,X):
        """
        Adds a new configuration.
        
        Args:
            conf (dictionary): The magnetic configuration: {atom number: magnetic moment}
            X: The tensor of that configuration
        """

        self.confs[self.nconfs] = conf
        self.Xs[self.nconfs] = X
        self.nconfs += 1
        
    def is_in(self,conf):
        #tests if a configuration is in
            
        out = False
        for i in range(self.nconfs):
            out_t = True
            for j in self.confs[i]:
                c = self.confs[i][j] - conf[j]
                for l in range(3):
                    if c[l].round(4)!= 0:
                        out_t = False
            if out_t:
                out = True

        return out

    def pprint(self,eo=None,m=-1,latex=False,print_format=None):
        #prints everything in a (somewhat) nice form
        #if m is set it prints only configuration m

        if m == -1:
            ran = list(range(self.nconfs))
        else:
            ran = [m]
        
        for n in ran:
            if n== 0:
                print('starting configuration:')
            else:
                print('configuration %s' %n)
            for p in sorted(self.confs[n].keys()):
                print('atom %s, m = %s, %s, %s' %(p,self.confs[n][p][0].round(4),self.confs[n][p][1].round(4),self.confs[n][p][2].round(4)))
                #print 'atom %s, m = %s, %s, %s' %(p,self.confs[n][p][0],self.confs[n][p][1],self.confs[n][p][2])
            print('First part of the response tensor')
            self.Xs[n][0].pprint(print_format=print_format)
            if latex:
                self.Xs[n][0].pprint(latex=True)
            print('Second part of the response tensor')
            self.Xs[n][1].pprint(print_format=print_format)
            if latex:
                self.Xs[n][1].pprint(latex=True)
            print('')

    def convert(self,T,shift):
        #converts to a different coordinate system
        #seems to work, but not fully tested, not sure if the shift will work correctly
        #not actually used anywhere, but may be useful in the future

        Tt = mat2ten(T)    

        confs_t = confs()
        for n in range(self.nconfs):

            conf_t = {}
            for j in self.confs[n]:
                pos = self.confs[n][j]
                pos_t = np.dot(T,pos) + np.array(shift)
                conf_t[j] = pos_t

            X = self.Xs[n]
            X_t = []
            X_t.append(Tt*X[0]*Tt.T())
            X_t.append(Tt*X[1]*Tt.T())

            confs_t.add(conf_t,X_t)

        return confs_t

def find_equiv(Xs,mag,syms,atom,debug=False,round_prec=None):
    """
    Takes a tensor and a list of nonmagnetic symmetries and find the form of the tensor for all equivalent configurations.

    !!!Will only work if the same input basis was used for the nonmagnetic and magnetic!!!

    Args:
        X: The tensor for the starting configuration.
        op1: First operator type.
        op2: Second operator type.
        atom: Sets a projection to an atom.
        syms: The nonmagnetic symmetry operations.
        mag: The magnetic moments - ie the starting configuration.
        T: is the transformation matrix from the nonmagnetic basis to the input basis

    Returns:
        C (confs class): Contains all the equivalent configurations and the transformed tensors for each.
    """
    
    if debug:
        print('starting find_equiv')

    n_at = []
    for sym in syms:
        n_at.append(len(sym.permutations))
    n_at = list(set(n_at))
    if len(n_at) != 1:
        raise Exception('Different number of atoms permutations for different symmetries. Something wrong.')
    else:
        n_at = n_at[0]

    #extracts the starting configuration, only the magnetic moments are needed
    start_conf = OrderedDict()
    for i in range(len(mag)):
        a = mag[i][0]
        b = mag[i][1]
        c = mag[i][2]
        if a.round(4) !=0 or b.round(6) !=0 or c.round(6) != 0:
            start_conf[i+1] = sympy.Matrix([[a,b,c]])

    if n_at != len(start_conf):
        message = ('Warning: nonmagnetic unit cell smaller than the magnetic!\n'
                  'Output will show values of only some magnetic moments.'
                  'Furthermore some equivalent configurations will be missing.'
                  'Use with caution, this is not properly tested.')
        print('')
        print(message)
        start_conf2 = OrderedDict()
        for p in start_conf:
            if p in syms[0].permutations:
                start_conf2[p] = start_conf[p]
        start_conf = start_conf2

    #creates a conf class, which stores all the configurations and adds the starting one
    C = confs()
    C.add(start_conf,Xs)

    if debug:
        C.pprint(0)

    for sym in syms: #a loop over all symmetries
        
        #if there is a projection take only the symmetry that keeps the atom invariant
        if atom == -1 or sym.permutations[atom] == atom:

            if debug:
                print('')
                print('taking sym : ')
                print(sym)
                print('')

            #transforms the starting configuration by the symmetry
            conf_t = OrderedDict()
            for p in start_conf:
                mom = sympy.Matrix([[start_conf[p][0],start_conf[p][1],start_conf[p][2]]])
                mom_T = sym.get_R('s') * mom.T
                mom_T = mom_T.T
                conf_t[sym.permutations[p]] = sympy.Matrix([[mom_T[0],mom_T[1],mom_T[2]]])

            if debug:
                print('symmetry transforms starting configuration to configuration:')
                for p in conf_t:
                    print('atom %s, m = %s, %s, %s' %(p,conf_t[p][0],conf_t[p][1],conf_t[p][2]))
            
            #if the configuration has not been found before it transforms the tensor by the symmetry
            #operation and adds everything to C
            if not C.is_in(conf_t):
                if debug:
                    print('configuration has not been found before')
                Xt = []
                Xt.append(Xs[0].transform(sym))
                Xt.append(Xs[1].transform(sym))
                if debug:
                    print('even part converted to:')
                    Xt[0].pprint()
                    print('odd part converted to:')
                    Xt[1].pprint()
                if round_prec is not None:
                    for X in Xt:
                        X.round(round_prec)
                C.add(conf_t,Xt)
    
    return C




from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .tensors import tensor

X = tensor('s',3,5)
X.pprint(print_format=2)



from __future__ import print_function
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import range
import re
import sys
import os

def find_group(name):

    dirname, filename = os.path.split(os.path.abspath(__file__))
    with open(dirname+'/findsym/mag_groups.txt') as f:
        lines = f.readlines()
    
    found = False
    for line in lines:
        line = line.split(',',1)
        name_text = line[0].split()[2]
        name_num = line[0].split()[1]
        if name == name_text or name == name_num:
            found = True
            break

    if not found:
        return None,None
    else:
        return name_text,name_num

def group_sym(name,debug=False):
    """Returns symmetry operations for a given group.

    Args:
        name: Name of the point group. Must be one of the magnetic point groups as given in the Sympy tables.
                Both number and name are accepted.
        dirname: Directory in which group tables are located.
                   Requires files syms_table.dat, syms_table_hex.dat and mag_groups.dat

    Returns:
        hex_group: Boolean. True if hexagonal or trigonal group, False otherwise. Useful for basis transformations.
        syms: The symmetrey operations in the format that is used in the code. The same format is outputted by the read.py module.
    """

    dirname, filename = os.path.split(os.path.abspath(__file__))

    with open(dirname+'/findsym/syms_table.dat') as f:
        lines = f.readlines()
        syms_t = {}
        for line in lines:
            syms_t[line.split()[0]] = (line.split()[1],line.split()[2])

    with open(dirname+'/findsym/syms_table_hex.dat') as f:
        lines = f.readlines()
        syms_t_hex = {}
        for line in lines:
            syms_t_hex[line.split()[0]] = (line.split()[1],line.split()[2])

    hex_syms = ['6z','3z','3z-1','6z-1','21','22','23','-6z','-3z','-3z-1','-6z-1','m1','m2','m3']

    with open(dirname+'/findsym/mag_groups.txt') as f:
        lines = f.readlines()
    
    found = False
    for line in lines:
        line = line.split(',',1)
        name_text = line[0].split()[2]
        name_num = line[0].split()[1]
        if name == name_text or name == name_num:
            line_f = line
            ops = [x[2:-1] for x in re.findall('\'\([12346xyzm\-]+\|',line[1])]
            ops_T = [x[2:-1] for x in re.findall('\"\([12346xyzm\-]+\|',line[1])]
            if debug:
                print('operations without time-reversal:', ops)
                print('operations with time-reversal:', ops_T)
            hex_group = False
            for op in ops + ops_T:
                if op in hex_syms:
                    hex_group = True

            sym_a=(name_text,name_num,hex_group,ops,ops_T)

            found = True
            break

    if not found:
        sys.exit('Group not found')
    syms=[]

    n = len(sym_a[3])
    for j in range(n):

        op = sym_a[3][j]

        if sym_a[2]:
            ops = syms_t_hex[op][0]
            opm = syms_t_hex[op][1]
        else:
            ops = syms_t[op][0]
            opm = syms_t[op][1]

        ops = ops.split(',')
        opm = opm.split(',')

        if debug:
            print('no time reversal')
            print('operator', op)
            print('symmetry operation:', [j,ops,opm,'+1'])
            print('')

        syms.append([j,ops,opm,'+1'])

    for j in range(len(ops_T)):

        op = sym_a[4][j]

        if sym_a[2]:

            ops = syms_t_hex['T'+op][0]
            opm = syms_t_hex['T'+op][1]
            ops = ops.split(',')
            opm = opm.split(',')

        else:

            ops = syms_t[op][0]
            opm = syms_t[op][1]
            ops = ops.split(',')
            opm = opm.split(',')

            for i in range(3):

                m = opm[i]

                if m == 'mx':
                    opm[i] = '-mx'
                if m == 'my':
                    opm[i] = '-my'
                if m == 'mz':
                    opm[i] = '-mz'
                if m == '-mx':
                    opm[i] = 'mx'
                if m == '-my':
                    opm[i] = 'my'
                if m == '-mz':
                    opm[i] = 'mz'

        if debug:
            print('time reversal')
            print('operator', op)
            print('symmetry operation:', [n+j,ops,opm,'-1'])
            print('')

        syms.append([n+j,ops,opm,'-1'])

    return hex_group,syms





from __future__ import division
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from past.utils import old_div
def convert_index(i,j):
    matrix = ['00','01','02','10','11','12','20','21','22']
    n = i*3+j
    return n

def convert_index_3(i,j,k):
    n = i*9 + j*3+k
    return n


# converts from 1d form index to a 3x3 matrix indeces
def inconvert_index(n):
    matrix = ['00','01','02','10','11','12','20','21','22']
    i=int(matrix[n][0])
    j=int(matrix[n][1])
    return [i,j]

def inconvert_index_3(n):
    i = old_div(n,9)
    j = old_div((n-i*9),3)
    k = n-i*9-j*3
    return i,j,k

def convert_index_rev(i,j):
    n = 8 - (i*3+j)
    return n

# converts from 1d form index to a 3x3 matrix indeces in a reversed order
def inconvert_index_rev(n):
    matrix = list(reversed(['00','01','02','10','11','12','20','21','22']))
    i=int(matrix[n][0])
    j=int(matrix[n][1])
    return [i,j]

def convert_index_rev_3(i,j,k):
    n = 26 - convert_index_3(i,j,k)
    return n

def inconvert_index_rev_3(n):
    n = 26 - n
    return inconvert_index_3(n)




from __future__ import print_function
from __future__ import absolute_import
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from builtins import str
from builtins import range
import re
import sys
import os

from . import symmetrize
from . import symmetrize_exp as st
from . import fslib
from . import find_eq
from . import symT
from . import mham
from . import groups
from .tensors import tensor,matrix
from . import symmetry

import sympy
from sympy import sympify as spf

#this finds the location of the main.py file and ads this location to the path where modules are searched
#this way the modules have to be present only in the install directory and not in the run directory
dirname, filename = os.path.split(os.path.abspath(__file__))
sys.path.append(str(dirname))

def def_symmetr_opt(opt):
    s_opt = symmetrize.SymmetrOpt(num_prec=opt['num_prec'],
                                   debug=opt['debug_sym'],
                                   debug_time=opt['debug_time'],
                                   debug_Y=opt['debug_symY'],
                                   round_prec=opt['round_prec'])
    return s_opt

def sym_res_nonexp(opt,printit=False):

    #the symmetry operations given in the basis used by findsym
    if opt['noso']:
        syms = symT.get_syms_noso(opt)
    else:
        syms = symT.get_syms(opt)

    #transformation matrix from the basis used by findsym to the user defined basis
    T = symT.get_T(opt)
    if opt['num_prec'] is not None:
        T = sympy.N(T)
    #if transform_result is not set we transform the symmetries
    if not opt['transform_result']:
        for sym in syms:
            sym.convert(T,in_place=True)

    op_contravar = (1,)*opt['op_lengths'][0] + (-1,)*opt['op_lengths'][1]

    op_types = opt['op_types']
    X1 = tensor('s',3,len(op_types),ind_types=op_contravar)
    X1.def_trans(ind_trans=op_types,T_comp=1)
    X2 = tensor('s',3,len(op_types),ind_types=op_contravar)
    X2.def_trans(ind_trans=op_types,T_comp=-1)

    eo = symmetrize.even_odd([X1,X2])

    same_op_sym = False
    if opt['same_op_sym']:
        if len(set(op_types)) == 1:
            same_op_sym = True
            #the metric is for the findsym basis
            G = symT.get_metric(opt)
            if opt['num_prec'] is not None:
                G = sympy.N(G)
            if not opt['transform_result']:
                G = T * G * T.T

            X1.def_metric(G)
            X2.def_metric(G)

    s_opt = def_symmetr_opt(opt)

    Xs = []
    Xs.append(symmetrize.symmetrize_res(syms,X1,opt['atom'],s_opt))
    Xs.append(symmetrize.symmetrize_res(syms,X2,opt['atom'],s_opt))

    if same_op_sym:
        for i in range(2):
            Xs[i] = symmetrize.symmetrize_same_op(Xs[i],s_opt)

    if opt['atom2'] != -1:
        Xs_2 = symmetrize.symmetr_AB(syms,Xs,opt['atom'],opt['atom2'],round_prec=opt['round_prec'])

    #if transform result is set we convert the symmetrized tensor
    if opt['transform_result']:
        for i in range(2):
            Xs[i].convert(T)
        if opt['atom2'] != -1:
            for i in range(2):
                Xs_2[i].convert(T)


    if printit:
        print('{0} part of the response tensor:'.format(eo[0]))
        Xs[0].pprint(print_format=opt['print_format'])
        if opt['latex']:
            Xs[0].pprint(print_format=opt['print_format'],latex=True)
        print('{0} part of the response tensor:'.format(eo[1]))
        Xs[1].pprint(print_format=opt['print_format'])
        if opt['latex']:
            Xs[1].pprint(print_format=opt['print_format'],latex=True)

        
        if opt['atom2'] != -1:
            if Xs_2 is None:
                print('no relation with atom %s found' % opt['atom2'])
            else:
                print('First part of the response tensor, atom %s' % (opt['atom2']))
                Xs_2[0].pprint(print_format=opt['print_format'])
                if opt['latex']:
                    Xs_2[0].pprint(print_format=opt['print_format'],latex=True)
                print('Second part of the response tensor, atom %s' % (opt['atom2']))
                Xs_2[1].pprint(print_format=opt['print_format'])
                if opt['latex']:
                    Xs_2[1].pprint(print_format=opt['print_format'],latex=True)
                print('')
    
    if opt['equiv']:

        mags = symT.get_mags(opt['inp']) #magnetic moments 
        syms_nm = symT.get_syms_nonmag(opt) #the symmetry operations for nonmagnetic system

        Tm = symT.get_Tm(opt['inp']) #transformation matrix from the input basis to the basis used by findsym
        #transformation matrix form the input basis to the basis used by findsym for nonmagnetic system 
        Tnm = symT.get_Tm(opt['inp'],nonmag=True) 

        #magnetic moments are in the input basis, we need to convert them to the user selected basis
        #First convert to the magnetic findsym basis and from that to the user selected one.
        mags = symT.convert_vecs(mags,T*Tm.inv())
        #The symmetry operations are in the findsym uses for nonmagnetic input. We first transform to the input
        #basis, then to the magnetic findsym basis and then to the user chosen one.
        T2 = T*Tm.inv()*Tnm
        for sym in syms_nm:
            sym.convert(T2,in_place=True)

        C = find_eq.find_equiv(Xs,mags,syms_nm,opt['atom'],debug=opt['debug_equiv'],round_prec=opt['round_prec'])
        if printit:
            print('')
            print('Equivalent configurations:')
            C.pprint(print_format=opt['print_format'])
    
    if opt['atom2'] == -1:
        if opt['equiv']:
            return Xs,C
        else:
            return Xs
    else:
        return Xs,Xs_2

def sym_res_exp(opt,printit=False):
    """
    Finds the tensor describing expansion term of linear response in the direction of magnetization.

    Args:
        opt (class options): stores all the input arguments. Only some are used here.
        printit (optional[boolean]): if true this prints the output
    """

    if opt['group']:
        print('!!!The input group must be one of the nonmagnetic point groups, otherwise the ouput will be wrong.!!!') 
        syms_nm = symT.get_syms(opt)
        T = symT.get_T(opt)
        mags = []
    else:
        T = symT.get_T(opt,nonmag=True)
        syms_nm = symT.get_syms_nonmag(opt)
        mags = symT.get_mags(opt['inp'])

    if not opt['transform_result']:
        for sym in syms_nm:
            sym.convert(T,in_place=True)
        mags = symT.convert_vecs(mags,T)

    syms_L = st.def_syms_L(mags,syms_nm,debug=False)

    op_contravar = (1,)*opt['op_lengths'][0] + (-1,)*opt['op_lengths'][1] + (-1,) * opt['exp']
    op_types = opt['op_types'] + ['L'] * opt['exp']
    X = tensor('s',3,len(op_types),ind_types=op_contravar)
    X.def_trans(ind_trans=op_types,T_comp=1)
    T_inv = symmetry.create_T()
    T_inv.def_custom_R('L',T_inv.Rs)
    if not X.is_even(T_inv):
        X.def_trans(ind_trans=op_types,T_comp=-1)
    s_opt = def_symmetr_opt(opt)

    X = symmetrize.symmetrize_res(syms_L,X,opt['atom'],s_opt)
    
    if opt['transform_result']:
        X.convert(T)


    if printit:
        n_op = opt['op_lengths'][0] + opt['op_lengths'][1] 
        st.print_tensor(X,n_op)
        if opt['latex']:
          st.print_tensor(X,latex=opt['latex'])

    return X

def sym_res(opt,printit=False):
    """A wrapper function that returns the appropriate response tensor based on the input options.

    Args:
        opt (class options): stores all the input arguments. Only some are used here.
        printit (optional[boolean]): if true this prints the output
    """
    if printit and opt['group'] is not None:
        group_name,group_num = groups.find_group(opt['group'])
        print('group name: ', group_name, 'group number: ', group_num)
    if opt['exp'] == -1:
        return sym_res_nonexp(opt,printit=printit)
    else:
        return sym_res_exp(opt,printit=printit)

def sym_mham(opt,printit=False):
    T = symT.get_T(opt,nonmag=True)
    syms = symT.get_syms_nonmag(opt)

    if not opt['transform_result']:
        for sym in syms:
            sym.convert(T,in_place=True)

    s_opt = def_symmetr_opt(opt)
    H = mham.sym_mag_ham(opt['sites'],syms,None,s_opt)

    if opt['transform_result']:
        #H = mham.convert_mag_ham(H,T)
        H.convert(T)

    if opt['equiv']:
        H_E = mham.equiv(H,opt['sites'],syms,T)

    if printit:
        if H.dim2 == 2:
            print('Hamiltonian term in matrix form:')
            H.pprint(latex=opt['latex'])
            print('')
        mham.print_Ham(H,opt['sites'],latex=opt['latex'])
        if opt['equiv']:
            print('')
            print('Hamiltonian terms for all equivalent combinations of sites:')
            for sites in H_E:
                print(str(sites)+':')
                mham.print_Ham(H_E[sites],sites,latex=opt['latex'])
                print('')
    if not opt['equiv']:
        return H
    else:
        return H,H_E

